(******************************************************************************)
(* generator (stack machine)                                                  *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Gen;

FROM SYSTEM IMPORT ADR;
IMPORT Lib, FIO;
IMPORT Comp;

CONST
  err ::= FIO.ErrorOutput;

TYPE
  TSet = SET OF [ 0 .. 31 ];

VAR
  op0s: ARRAY TOp0 OF Comp.TOperation;
  op1s: ARRAY TOp1 OF Comp.TOperation;
  op2s: ARRAY TOp2 OF Comp.TOperation;
  xops: ARRAY TOp2 OF Comp.TOperation;
  vals: ARRAY BOOLEAN OF TValue;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Gen: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;


PROCEDURE Val( v: TValue );
BEGIN
  Comp.Val( v );
END Val;

PROCEDURE Sto( i: TIndex );
BEGIN
  Comp.Sto( i );
END Sto;

PROCEDURE Lod( i: TIndex );
BEGIN
  Comp.Lod( i );
END Lod;

PROCEDURE Op0( o: TOp0 );
BEGIN
  Comp.Opr( op0s[o] );
END Op0;

PROCEDURE Op1( o: TOp1; VAR d: TItemDesc );
BEGIN
  IF d.c THEN
    CASE o OF
    | oNeg: d.v := -d.v;
    | oNot: d.v := TValue( TSet(d.v) / TSet(TValue(-1)) );
    ELSE
      Err( 'Bug: Unknown unary operator' );
    END;
  ELSE
    Comp.Opr( op1s[o] );
  END;
END Op1;

PROCEDURE Op2( o: TOp2; VAR d, d1: TItemDesc );
BEGIN
  IF d.c & d1.c THEN
    CASE o OF
    | oAdd: d.v := d.v + d1.v;
    | oSub: d.v := d.v - d1.v;
    | oMul: d.v := d.v  *  d1.v;
    | oDiv: d.v := d.v DIV d1.v;
    | oOr : d.v := TValue( TSet(d.v) + TSet(d1.v) );
    | oAnd: d.v := TValue( TSet(d.v) * TSet(d1.v) );
    | oEql: d.v := vals[ d.v =  d1.v ];
    | oNeq: d.v := vals[ d.v #  d1.v ];
    | oLss: d.v := vals[ d.v <  d1.v ];
    | oLeq: d.v := vals[ d.v <= d1.v ];
    | oGtr: d.v := vals[ d.v >  d1.v ];
    | oGeq: d.v := vals[ d.v >= d1.v ];
    ELSE
      Err( 'Bug: Unknown binary operator' );
    END;
  ELSE
    IF d.c THEN
      Comp.Val( d.v );
      Comp.Opr( xops[o] );
    ELSE
      IF d1.c THEN Comp.Val( d1.v ) END;
      Comp.Opr( op2s[o] );
    END;
    d.c := FALSE;
  END;
END Op2;


PROCEDURE InitModule();
BEGIN
  Lib.Fill( ADR(op0s), SIZE(op0s), 0C );
  op0s[ oWrt ] := Comp.oWrt;
  op0s[ oWrB ] := Comp.oWrB;

  Lib.Fill( ADR(op1s), SIZE(op1s), 0C );
  op1s[ oNeg ] := Comp.oNeg;
  op1s[ oNot ] := Comp.oNot;

  Lib.Fill( ADR(op2s), SIZE(op2s), 0C );
  op2s[ oAdd ] := Comp.oAdd;
  op2s[ oSub ] := Comp.oSub;
  op2s[ oMul ] := Comp.oMul;
  op2s[ oDiv ] := Comp.oDiv;
  op2s[ oOr  ] := Comp.oOr;
  op2s[ oAnd ] := Comp.oAnd;
  op2s[ oEql ] := Comp.oEql;
  op2s[ oNeq ] := Comp.oNeq;
  op2s[ oLss ] := Comp.oLss;
  op2s[ oLeq ] := Comp.oLeq;
  op2s[ oGtr ] := Comp.oGtr;
  op2s[ oGeq ] := Comp.oGeq;

  Lib.Fill( ADR(xops), SIZE(xops), 0C );
  xops[ oAdd ] := Comp.oAdd;
  xops[ oSub ] := Comp.oBus;
  xops[ oMul ] := Comp.oMul;
  xops[ oDiv ] := Comp.oVid;
  xops[ oOr  ] := Comp.oOr;
  xops[ oAnd ] := Comp.oAnd;
  xops[ oEql ] := Comp.oEql;
  xops[ oNeq ] := Comp.oNeq;
  xops[ oLss ] := Comp.oGtr;
  xops[ oLeq ] := Comp.oGeq;
  xops[ oGtr ] := Comp.oLss;
  xops[ oGeq ] := Comp.oLeq;

  Lib.Fill( ADR(vals), SIZE(vals), 0C );
  vals[ FALSE ] := Comp.vFalse;
  vals[ TRUE  ] := Comp.vTrue;
END InitModule;

BEGIN
  InitModule();
END Gen.
