(******************************************************************************)
(* plx parser                                                                 *)
(* (C) 2010 Steffen Solyga <solyga@absinth.net>                               *)
(******************************************************************************)
IMPLEMENTATION MODULE Prsr;

IMPORT FIO;
FROM Scnr IMPORT Put, Read, KeepId, Diff, TSymbol, sym, num, id, pos;
FROM Gen IMPORT TValue, TOp0, TOp1, TOp2, TItemDesc;
FROM Gen IMPORT Val, Sto, Lod, Op0, Op1, Op2, vFalse, vTrue;

CONST
  err ::= FIO.ErrorOutput;
  cntMax = 32;

TYPE
  TClass = ( cConst, cVar, cType );
  TType = ( tBool, tInt );
  TObjDesc = RECORD
    id : CARDINAL;
    typ: TType;
    CASE cls: TClass OF
    | cConst: val: TValue;
    | cVar  : idx: CARDINAL;
    | cType :
    END;
  END;

VAR
  objs: ARRAY [ 0 .. cntMax - 1 ] OF TObjDesc;
  cnt: CARDINAL;
  idx: CARDINAL;
  op2s: ARRAY TSymbol OF TOp2;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Prsr (' );
  FIO.WrCard( err, pos.l, 0 );
  FIO.WrChar( err, ',' );
  FIO.WrCard( err, pos.c, 0 );
  FIO.WrStr( err, '): ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Find( VAR i: CARDINAL );
BEGIN
  i := 0;
  WHILE i < cnt DO
    IF Diff( objs[i].id, id ) = 0 THEN RETURN END;
    INC( i );
  END;
END Find;


PROCEDURE Expression( VAR t: TType; VAR d: TItemDesc );

  PROCEDURE SimpleExpr( VAR t: TType; VAR d: TItemDesc );

    PROCEDURE Factor( VAR t: TType; VAR d: TItemDesc );
    VAR i: CARDINAL;
    BEGIN
      CASE sym OF
      | sNum : IF num > MAX( TValue ) THEN Err( 'Number too large' ) END;
               t := tInt; d.c := TRUE; d.v := num;
               Read();
      | sId  : Find( i );
               IF i >= cnt THEN Err( 'Identifier undeclared' ) END;
               t := objs[ i ].typ;
               CASE objs[ i ].cls OF
               | cConst: d.c := TRUE; d.v := objs[ i ].val;
               | cVar  : d.c := FALSE; Lod( objs[i].idx );
               ELSE
                 Err( 'Expected constant or variable' );
               END;
               Read();
      | sLPar: Read(); Expression( t, d );
               IF sym # sRPar THEN Err( 'Expected operator or ")"' ) END;
               Read();
      | sNot : Read(); Factor( t, d );
               IF t # tBool THEN Err( 'Expected boolean factor' ) END;
               Op1( oNot, d );
      ELSE
        Err( 'Expected factor' );
      END;
    END Factor;

    PROCEDURE Term( VAR t: TType; VAR d: TItemDesc );
    VAR s: TSymbol; t1: TType; d1: TItemDesc;
    BEGIN
      Factor( t, d );
      WHILE ( sMul <= sym ) & ( sym <= sAnd ) DO
        IF ( sym = sAnd ) # ( t = tBool ) THEN
          Err( 'Wrong operator type' )
        END;
        s := sym; Read(); Factor( t1, d1 );
        IF t # t1 THEN Err( 'Wrong operand type' ) END;
        Op2( op2s[s], d, d1 );
      END;
    END Term;

  VAR s: TSymbol; t1: TType; d1: TItemDesc;
  BEGIN
    IF ( sym = sPlus ) OR ( sym = sMinus ) THEN
      s := sym; Read(); Term( t, d );
      IF t # tInt THEN Err( 'Expected integer term' ) END;
      IF s = sMinus THEN Op1( oNeg, d ) END;
    ELSE
      Term( t, d );
    END;
    WHILE ( sPlus <= sym ) & ( sym <= sOr ) DO
      IF ( sym = sOr ) # ( t = tBool ) THEN Err( 'Wrong operator type' ) END;
      s := sym; Read(); Term( t1, d1 );
      IF t # t1 THEN Err( 'Wrong operand type' ) END;
      Op2( op2s[s], d, d1 );
    END;
  END SimpleExpr;

VAR s: TSymbol; t1: TType; d1: TItemDesc;
BEGIN
  SimpleExpr( t, d );
  IF ( sEql <= sym ) & ( sym <= sGeq ) THEN
    s := sym; Read(); SimpleExpr( t1, d1 );
    IF t # t1 THEN Err( 'Wrong operand type' ) END;
    Op2( op2s[s], d, d1 );
    t := tBool;
  END;
END Expression;

PROCEDURE Statement();
VAR i: CARDINAL; t: TType; d: TItemDesc;
BEGIN
  CASE sym OF
  | sId : (* assignment *)
          Find( i );
          IF i >= cnt THEN Err( 'Identifier undeclared' ) END;
          IF objs[ i ].cls # cVar THEN Err( 'Not a variable' ) END;
          Read();
          IF sym # sBec THEN Err( 'Expected ":="' ) END;
          Read(); Expression( t, d );
          IF t # objs[i].typ THEN Err( 'Incompatible types' ) END;
          IF d.c THEN Val( d.v ) END;
          Sto( objs[i].idx );
  | sWrt: (* expression output *)
          Read(); Expression( t, d );
          IF d.c THEN Val( d.v ) END;
          CASE t OF
          | tInt : Op0( oWrt );
          | tBool: Op0( oWrB );
          ELSE
            Err( 'Bug: Invalid type' );
          END;
  ELSE
    (* empty *)
  END;
END Statement;

PROCEDURE ConstDeclaration();
VAR i: CARDINAL; t: TType; d: TItemDesc;
BEGIN  (* sym = sId *)
  Find( i );
  IF i < cnt THEN Err( 'Identifier redeclared' ) END;
  IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
  objs[ i ].id  := id;
  objs[ i ].cls := cConst;
  KeepId();
  Read();
  IF sym # sEql THEN Err( 'Expected "="' ) END;
  Read(); Expression( t, d );
  IF ~d.c THEN Err( 'Expected constant' ) END;
  objs[ i ].typ := t;
  objs[ i ].val := d.v;
  INC( cnt );
  IF sym # sSemi THEN Err( 'Expected ";"' ) END;
  Read();
END ConstDeclaration;

PROCEDURE VarDeclaration();
VAR i,j: CARDINAL;
BEGIN  (* sym = sId *)
  Find( i );
  IF i < cnt THEN Err( 'Identifier redeclared' ) END;
  IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
  objs[ i ].id  := id;
  objs[ i ].cls := cVar;
  KeepId();
  Read();
  IF sym # sCol THEN Err( 'Expected ":"' ) END;
  Read();
  IF sym # sId THEN Err( 'Expected identifier' ) END;
  Find( j );
  IF j >= cnt THEN Err( 'Identifier undeclared' ) END;
  IF objs[ j ].cls # cType THEN Err( 'Not a type' ) END;
  objs[ i ].typ := objs[ j ].typ;
  objs[ i ].idx := idx;
  INC( cnt ); INC( idx );
  Read();
  IF sym # sSemi THEN Err( 'Expected ";"' ) END;
  Read();
END VarDeclaration;

PROCEDURE Run(); (* Program() *)
BEGIN
  Read();
  IF sym = sConst THEN
    Read();
    WHILE sym = sId DO ConstDeclaration() END;
  END;
  IF sym = sVar THEN
    Read();
    WHILE sym = sId DO VarDeclaration() END;
  END;
  IF sym # sBeg THEN Err( 'Expected "BEGIN"' ) END;
  Read(); Statement();
  WHILE sym = sSemi DO Read(); Statement() END;
  IF sym # sDot THEN Err( 'Expected ";" or "."' ) END;
(*  Read(); IF sym # sEOF THEN Err( 'Expected EOF' ) END; *)
END Run;

PROCEDURE InitOprs();
VAR s: TSymbol;
BEGIN
  FOR s := MIN( TSymbol ) TO MAX( TSymbol ) DO op2s[ s ] := oNull END;
  op2s[ sMul   ] := oMul;
  op2s[ sDiv   ] := oDiv;
  op2s[ sAnd   ] := oAnd;
  op2s[ sPlus  ] := oAdd;
  op2s[ sMinus ] := oSub;
  op2s[ sOr    ] := oOr;
  op2s[ sEql   ] := oEql;
  op2s[ sNeq   ] := oNeq;
  op2s[ sLss   ] := oLss;
  op2s[ sLeq   ] := oLeq;
  op2s[ sGtr   ] := oGtr;
  op2s[ sGeq   ] := oGeq;
END InitOprs;

BEGIN
  cnt := 0;
  idx := 0;
  Put( 'FALSE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := vFalse;
  KeepId(); INC( cnt );
  Put( 'TRUE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := vTrue;
  KeepId(); INC( cnt );
  Put( 'BOOLEAN' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cType;
  KeepId(); INC( cnt );
  Put( 'INTEGER' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tInt;
  objs[ cnt ].cls := cType;
  KeepId(); INC( cnt );
  InitOprs();
END Prsr.
